from base64 import b64encode
from typing import Optional, List, Callable
from aiohttp import web, ClientSession
from datetime import datetime, timezone
from pathlib import Path

from .base_auth_flow import BaseAuthFlow
from .token import Token
from ..exceptions import SpotifyException

import urllib.parse
import re
import secrets
import asyncio
import webbrowser


class OauthFlow(BaseAuthFlow):
    def __init__(
            self,
            client_id: str,
            client_secret: str,
            redirect_uri: str,
            # `scope` and `show_dialog` are None on purpose so that
            # the wrapper method removes them if not given
            scope: Optional[List[str]] = None,
            show_dialog: Optional[bool] = None,
            token: Optional[Token] = None,
            open_browser: bool = False,
            disable_builtin_server: bool = False,
            authenticated_html_response: Optional[str] = Path(__file__).parent / "./static/authenticated.html",
            failed_html_response: Optional[str] = Path(__file__).parent / "./static/failed.html",
            state_verifier: Optional[Callable[[str], bool]] = None
    ) -> None:
        """
        ### Args
        - client_id: `str`, Spotify client id
        - client_secret: `str`, Spotify client secret
        - redirect_uri: `str`, URI for Oauth callback, URI(s) using localhost (127.0.0.1)
        will automatically start spotipy2's built-in server. URI must match one of the URI's on
        the project's Spotify Dashboard character for character.
        - scope: `Optional[List[str]]`, List of Spotify scopes,
        full list: https://developer.spotify.com/documentation/general/guides/authorization/scopes/
        - show_dialog: `Optional[bool]`, Whether to force the user to approve the app again if they have already done so
        - token: `Optional[str]`, The access token
        - open_browser: `bool`, Whether to open the browser or not when get_redirect is called
        - disable_builtin_server: `bool`, If True, the built-in server will not be started
        - authenticated_html_response: `Optional[str]`, The path to a html file which will be displayed if a user is
        successfully authenticated, only works when using spotipy2's built-in server
        - failed_html_response: `Optional[str]`, The path to a html file which will be displayed if a user is
        NOT successfully authenticated, only works when using spotipy2's built-in server
        - state_verifier: `Optional[Callable[[str], bool]]`, Method to be executed to verify state to prevent XSS attack. 

        ### Returns
        - `None`

        ### Errors raised
        - None

        ### Function / Notes
        - For spotipy2's built-in server to start, the redirect_uri must contain localhost as
        the domain as well as the port, eg. http://localhost:9000. No path should be specified and HTTPS cannot be used
        and will be converted to HTTP. If you would like to use HTTPS, consider setting up your own web server.
        In addition, you must call the get_redirect method before executing any requests
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.scope = " ".join(scope) if scope else None
        self.show_dialog = str(show_dialog).lower() if show_dialog is not None else None

        self.token = token
        self.open_browser = open_browser
        self.authenticated_html_response = authenticated_html_response
        self.failed_html_response = failed_html_response
        self.disable_builtin_server = disable_builtin_server
        self.header = self.make_auth_header()

        self.state_verifier = state_verifier if state_verifier else self._state_verifier
        self.expected_state = secrets.token_urlsafe(16)

        self.server = None  # Oauth server
        self.server_runner = None
        self.timeout = 3

    async def get_redirect(self) -> str:
        """
        ### Args
        - None

        ### Returns
        - `str`, The authentication URI which the user must navigate to in order to get authenticated

        ### Errors raised
        - None

        ### Function / Notes
        - This will automatically start spotipy2's built-in server if the redirect_uri contains localhost as
        the domain as well as the port, eg. http://localhost:9000.
        """
        API_URL = "https://accounts.spotify.com/authorize?"

        data = self.wrapper(
            response_type="code",
            client_id=self.client_id,
            scope=self.scope,
            redirect_uri=self.redirect_uri,
            show_dialog=self.show_dialog,
            state=self.expected_state,
        )

        final_url = API_URL + urllib.parse.urlencode(data)

        if not self.disable_builtin_server:
            self.redirect_uri = self.redirect_uri.replace("https://", "http://")
            regex = '(?:http.*://)?(?P<host>[^:/ ]+).?(?P<port>[0-9]*).*'
            result = re.search(regex, self.redirect_uri)
            if result.group("host") in ['localhost', '127.0.0.1'] and result.group("port") != '':
                self.server = OauthServer(self.authenticated_html_response,
                                          self.failed_html_response,
                                          self.state_verifier)
                runner = web.AppRunner(self.server)
                await runner.setup()
                self.server_runner = web.TCPSite(runner, 'localhost', int(result.group("port")))
                # We don't need to implement any code for graceful shutdown as we don't do anything such as data
                # streaming
                await self.server_runner.start()
        
        if self.open_browser:
            webbrowser.open_new(final_url)

        return final_url

    async def get_access_token(self, http: ClientSession) -> Token:
        """
        ### Args
        - http: `ClientSession`, Aiohttp ClientSession to use

        ### Returns
        - `Token`, The access token

        ### Errors raised
        - None

        ### Function / Notes
        - None
        """
        if self.token and self.token.expires_at > datetime.now(timezone.utc):
            return self.token
        elif self.token and self.token.expires_at < datetime.now(timezone.utc):
            return await self.refresh_token(http)
        if not self.disable_builtin_server:
            while not self.server.code:
                await asyncio.sleep(self.timeout)
            code = self.server.code
        else:
            url = input("Please paste the URI you were redirected to: ")
            processed_url = urllib.parse.parse_qs(urllib.parse.urlparse(url).query)
            code = processed_url.get('code')
            state = processed_url.get('state')[0]
            if code:
                if await self._state_verifier(state):
                    code = code[0]
                else:
                    raise RuntimeError(f"State does not match. State: {state}")
            else:
                raise SpotifyException(403, f"Authentication failed. State: {state}")
        self.token = await self._get_access_token(code,
                                                  self.client_id,
                                                  self.client_secret,
                                                  self.redirect_uri,
                                                  http)
        return self.token

    @staticmethod
    async def _get_access_token(token: str, client_id: str, client_secret: str, redirect_uri: str,
                                http: ClientSession) -> Token:
        """
        ### Args
        - token: `str`, 'code' returned from Oauth callback (redirect_uri), this is NOT the access token
        - client_id: `str`, Spotify client id
        - client_secret: `str`, Spotify client secret
        - redirect_uri: `str`, URI for Oauth callback, URI must match one of the URI's on
        the project's Spotify Dashboard character for character
        - http: `ClientSession`, Aiohttp ClientSession to use

        ### Returns
        - `Token`, The access token

        ### Errors raised
        - None

        ### Function / Notes
        - This method is static so that developers who have already gotten the 'code' through their own code (for eg.
        they already have a separate website using flask) can easily use this method and instantiate a new auth
        flow with the access token which is returned
        """

        API_URL = "https://accounts.spotify.com/api/token"
        data = {
            "code": token,
            "redirect_uri": redirect_uri,
            "grant_type": "authorization_code"
        }

        HEADER = {
            "Authorization": "Basic %s"
                             % b64encode(f"{client_id}:{client_secret}".encode()).decode()
        }

        async with http.post(
                API_URL, data=data, headers=HEADER
        ) as r:
            if r.status < 200 or r.status > 299:
                raise SpotifyException(r.status, r.json()) # example: {'error': 'invalid_grant', 'error_description': 'Invalid redirect URI'}
            return await Token.from_dict(await r.json())

    async def refresh_token(self, http: ClientSession) -> Token:
        """
        ### Args
        - http: `ClientSession`, Aiohttp ClientSession to use

        ### Returns
        - `Token`, The access token

        ### Errors raised
        - None

        ### Function / Notes
        - None
        """
        API_URL = "https://accounts.spotify.com/api/token"
        data = {
            "refresh_token": self.token.refresh_token,
            "grant_type": "refresh_token"
        }

        HEADER = self.header

        async with http.post(
                API_URL, data=data, headers=HEADER
        ) as r:
            response = await r.json()
            # Refresh token is not received on token refresh
            response["refresh_token"] = self.token.refresh_token
            self.token = await Token.from_dict(response)
            return self.token

    async def _state_verifier(self, state: str) -> bool:
        """
        ### Args
        - state: `str`, State to verify

        ### Returns
        - `bool`, True if it is correct, False if not

        ### Errors raised
        - None

        ### Function / Notes
        - This is spotipy2's built in state_verifier
        """
        return state == self.expected_state


class OauthServer(web.Application):
    def __init__(self, authenticated_html_response: str, failed_html_response: str, state_verifier: Callable):
        self.code = None  # The code is *not* the access token
        self.state = None

        self.authenticated_html_response = authenticated_html_response
        self.failed_html_response = failed_html_response
        self.state_verifier = state_verifier

        super().__init__()
        self.add_routes([web.get('/', self.authorized)])

    async def authorized(self, request):
        self.state = request.rel_url.query['state']

        if await self.state_verifier(self.state):
            self.code = request.rel_url.query['code']
            return web.FileResponse(self.authenticated_html_response)
        else:
            return web.FileResponse(self.failed_html_response)



from __future__ import annotations
from typing import List, Optional, Type
from dataclasses import dataclass

from spotipy2 import types


@dataclass
class Paging(types.BaseType):
    href: str
    items: List[Type[types.BaseType]]
    limit: int
    next: Optional[str]
    offset: int
    previous: Optional[str]
    total: int

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"Paging(len(items)={len(self.items)}, limit={self.limit}, offset={self.offset}, total={self.total})"



from __future__ import annotations
from typing import List
from dataclasses import dataclass, field

from spotipy2 import types


@dataclass
class Artist(types.BaseType):
    external_urls: dict
    href: str
    id: str
    name: str
    uri: str

    followers: dict = field(default=None)
    genres: List[str] = field(default=None)
    images: List[dict] = field(default=None)
    popularity: int = field(default=None)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Artist(name='{self.name}', id='{self.id}')"



from __future__ import annotations
from typing import List, Optional, Union, Dict

import spotipy2
from spotipy2.types import Paging, Album, Artist, Playlist, Track


class SearchMethods:
    TYPES_SUPPORTED = [Album, Artist, Playlist, Track]

    async def search(
        self: spotipy2.Spotify,  # type: ignore
        query: str,
        types: Union[TYPES_SUPPORTED, List[TYPES_SUPPORTED]],
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        include_external: Optional[str] = None,
    ) -> Union[Paging, Dict[Paging]]:
        types = types if isinstance(types, list) else [types]

        params = self.wrapper(
            market=market, limit=limit, offset=offset, include_external=include_external
        )

        return await self._get(
            "search",
            params={
                **{"query": query, "type": ",".join(t.__name__.lower() for t in types)},
                **params,
            },
        )



from __future__ import annotations
from typing import Optional, List
from dataclasses import dataclass, field

from spotipy2 import types


@dataclass
class Album(types.BaseType):
    album_type: str
    total_tracks: int
    artists: List[types.Artist]
    external_urls: dict
    href: str
    id: str
    images: List[dict]
    name: str
    release_date: str
    release_date_precision: str
    uri: str

    available_markets: Optional[List[str]] = field(default=None)
    copyrights: List[dict] = field(default=None)
    external_ids: dict = field(default=None)
    genres: List[str] = field(default=None)
    label: str = field(default=None)
    popularity: int = field(default=None)
    tracks: List[types.Track] = field(default=None)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Album(name='{self.name}', id='{self.id}')"



from .base_auth_flow import BaseAuthFlow
from .client_credentials_flow import ClientCredentialsFlow
from .oauth_flow import OauthFlow
from .token import Token

__all__ = ["BaseAuthFlow", "ClientCredentialsFlow", "Token", "OauthFlow"]



from __future__ import annotations
from datetime import datetime
from dataclasses import dataclass

from spotipy2 import types


@dataclass
class PlaylistItem(types.BaseType):
    added_at: datetime
    added_by: dict
    is_local: bool
    track: types.Track
    video_thumbnail: dict

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"PlaylistItem(track='{self.track.name}')"



from __future__ import annotations
from typing import List, Optional

import spotipy2
from spotipy2.types import Artist, Track, Paging


class ArtistMethods:
    async def get_artists(
        self: spotipy2.Spotify, artist_ids: List[str]  # type: ignore
    ) -> List[Artist]:
        artists = await self._get(
            "artists", params={"ids": ",".join([self.get_id(i) for i in artist_ids])}
        )
        return artists["artists"]

    async def get_artist(
        self: spotipy2.Spotify, artist_id: str  # type: ignore
    ) -> Artist:
        return await self._get(f"artists/{self.get_id(artist_id)}")

    async def get_artist_top_tracks(
        self: spotipy2.Spotify, artist_id: str, market: str  # type: ignore
    ) -> List[Track]:
        top_tracks = await self._get(
            f"artists/{self.get_id(artist_id)}/top-tracks", params={"market": market}
        )
        return top_tracks["tracks"]

    async def get_artist_albums(
        self: spotipy2.Spotify,  # type: ignore
        artist_id: str,
        include_groups: Optional[str] = None,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Paging:
        params = self.wrapper(
            include_groups=include_groups, market=market, limit=limit, offset=offset
        )

        return await self._get(
            f"artists/{self.get_id(artist_id)}/albums", params=params
        )



from __future__ import annotations
from dataclasses import dataclass, field

from spotipy2 import types


@dataclass
class User(types.BaseType):
    href: str
    id: str
    uri: str

    display_name: str = field(default=None)
    external_urls: dict = field(default=None)
    followers: dict = field(default=None)
    images: dict = field(default=None)

    def __str__(self) -> str:
        return self.display_name

    def __repr__(self) -> str:
        if self.display_name:
            return f"User(display_name='{self.display_name}', id='{self.id}')"
        else:
            return f"User(id='{self.id}')"



from __future__ import annotations
from typing import List, Optional

import spotipy2
from spotipy2.types import Album
from spotipy2.types.paging import Paging


class AlbumMethods:
    async def get_albums(
        self: spotipy2.Spotify, album_ids: List[str]  # type: ignore
    ) -> List[Album]:
        albums = await self._get(
            "albums", params={"ids": ",".join([self.get_id(i) for i in album_ids])}
        )
        return albums["albums"]

    async def get_album(self: spotipy2.Spotify, album_id: str) -> Album:  # type: ignore
        return await self._get(f"albums/{self.get_id(album_id)}")

    async def get_album_tracks(
        self: spotipy2.Spotify,  # type: ignore
        album_id: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Paging:
        params = self.wrapper(market=market, limit=limit, offset=offset)

        album_tracks = await self._get(
            f"albums/{self.get_id(album_id)}/tracks", params=params
        )
        return album_tracks



from __future__ import annotations
from typing import List
from dataclasses import dataclass

from spotipy2 import types


@dataclass
class Playlist(types.BaseType):
    collaborative: bool
    description: str
    external_urls: dict
    followers: dict
    href: str
    id: str
    images: List[dict]
    name: str
    owner: dict
    public: bool
    snapshot_id: str
    tracks: List[types.Track]
    uri: str

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Playlist(name='{self.name}', id='{self.id}')"



from __future__ import annotations
from typing import Type
import inspect


class BaseType:
    @classmethod
    def from_dict(cls, json: dict) -> Type[BaseType]:
        if "_id" in json:  # MongoDB _id parameter
            json.pop("_id")

        try:
            return cls(**json)
        except TypeError:
            # Remove unwanted parameter(s)
            class_keys = inspect.signature(cls).parameters.keys()
            new_json = {k: v for k, v in json.items() if k in class_keys}

            # Return the new class without any extra parameter
            return cls(**new_json)



from base64 import b64encode


class BaseAuthFlow:
    def make_auth_header(self) -> dict:
        return {
            "Authorization": "Basic %s"
                             % b64encode(f"{self.client_id}:{self.client_secret}".encode()).decode()
        }

    @staticmethod
    def wrapper(**kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}



import re

from .albums import AlbumMethods
from .artists import ArtistMethods
from .search import SearchMethods
from .tracks import TrackMethods
from .playlists import PlaylistMethods


class Methods(
    AlbumMethods, ArtistMethods, SearchMethods, TrackMethods, PlaylistMethods
):
    @staticmethod
    def get_id(s: str) -> str:
        if m := re.search("(?!.*/).+", s):
            return m[0].split("?")[0]
        else:
            raise ValueError("ID not valid")

    @staticmethod
    def wrapper(**kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}



from .base import BaseType
from .paging import Paging
from .album import Album
from .artist import Artist
from .playlist import Playlist
from .track import Track
from .user import User
from .playlist_item import PlaylistItem

SPOTIFY = {
    "album": Album,
    "artist": Artist,
    "playlist": Playlist,
    "track": Track,
    "user": User
}

__all__ = [
    "BaseType",
    "Paging",
    "Album",
    "Artist",
    "Playlist",
    "Track",
    "User",
    "PlaylistItem"
]



from __future__ import annotations
from typing import List

import spotipy2
from spotipy2.types import Track


class TrackMethods:
    async def get_tracks(
        self: spotipy2.Spotify, track_ids: List[str]  # type: ignore
    ) -> List[Track]:
        r = await self._get(
            "tracks",
            params={"ids": ",".join([self.get_id(i) for i in track_ids])}
        )

        return r["tracks"]

    async def get_track(self: spotipy2.Spotify, track_id: str) -> Track:  # type: ignore
        return await self._get(f"tracks/{self.get_id(track_id)}")



from __future__ import annotations
from typing import Optional, List
from dataclasses import dataclass, field

from spotipy2 import types


@dataclass
class Track(types.BaseType):
    artists: List[types.Artist]
    disc_number: int
    duration_ms: int
    explicit: bool
    external_urls: dict
    href: str
    id: str
    is_local: bool
    name: str
    track_number: int
    uri: str

    album: Optional[types.Album] = field(default=None)
    external_ids: Optional[dict] = field(default=None)
    popularity: Optional[int] = field(default=None)
    is_playable: Optional[bool] = field(default=None)
    preview_url: Optional[str] = field(default=None)
    available_markets: Optional[List[str]] = field(default=None)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Track(name='{self.name}', id='{self.id}')"



from __future__ import annotations
from typing import Optional, Union, Type

import re
import logging
import asyncio
import inspect
from aiohttp import ClientSession

import spotipy2.types as types
from spotipy2.auth import ClientCredentialsFlow, OauthFlow
from spotipy2.methods import Methods
from spotipy2.exceptions import SpotifyException


class Spotify(Methods):
    API_URL = "https://api.spotify.com/v1/"

    def __init__(
        self,
        auth_flow: Union[ClientCredentialsFlow, OauthFlow],
        mongodb_uri: Optional[str] = None,
        auto_conversion: bool = True,
        recursive_conversion: bool = True,
        *args,
        **kwargs,
    ) -> None:
        self.auth_flow = auth_flow
        self.http = ClientSession(*args, **kwargs)
        self.auto_conversion = auto_conversion
        self.recursive_conversion = recursive_conversion

        if mongodb_uri:
            from pymongo import MongoClient

            match = re.match(r"(mongodb:\/\/\S+\/)(\w+)(\?\S+)?", mongodb_uri)

            if match:
                db_url, db_name, db_params = match.groups()
            else:
                raise Exception("MongoDB URI Not valid")

            db = MongoClient(db_url + (db_params or ""))
            self.cache = db[db_name].spotipy2
        else:
            self.cache = None

    async def _req(
        self,
        method: str,
        endpoint: str,
        params: Optional[dict] = None,
        can_be_cached: bool = False,
    ) -> dict:
        cache_enabled = self.cache is not None and can_be_cached

        # Get cached element
        if cache_enabled:
            doc = self.cache.find_one({"_endpoint": endpoint})
            if doc:
                doc.pop("_endpoint")
                doc.pop("_id")
                return self.convert(doc) if self.auto_conversion else doc

        token = await self.auth_flow.get_access_token(self.http)
        headers = {"Authorization": f"Bearer {token.access_token}"}

        async with self.http.request(
            method, f"{self.API_URL}{endpoint}", params=params, headers=headers
        ) as r:
            json = await r.json()

            try:
                assert r.status == 200
            except AssertionError:
                raise SpotifyException(
                    json["error"]["status"], json["error"]["message"]
                )
            else:
                # Save in cache if enabled
                if cache_enabled:
                    asyncio.create_task(self.cache_resource(endpoint, json))

                return self.convert(json) if self.auto_conversion else json

    async def _get(self, endpoint: str, params: Optional[dict] = None) -> dict:
        # Check if cache is enabled and request is a simple get [resource]
        can_be_cached = self.cache is not None and (
            params is None
            and re.match(r"^(?!me|browse)([\w-]+)\/(\w+)$", endpoint) is not None
        )

        return await self._req("GET", endpoint, params, can_be_cached)

    async def stop(self) -> None:
        await self.http.close()

    async def __aenter__(self) -> Spotify:
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.stop()

    async def cache_resource(self, endpoint, value) -> None:
        try:
            # Insert endpoint for future requests
            value["_endpoint"] = endpoint
            self.cache.insert_one(value)
        except Exception as e:
            logging.warning(f"Can't insert cached object: {e}")

    def convert(self, json: str, recursive: Optional[bool] = None):
        if recursive is None:
            recursive = self.recursive_conversion

        # Find and convert sub-items (recursive by default)
        if recursive:
            list_args = json.items() if isinstance(json, dict) else enumerate(json)
            convertible_args = {
                k: v for k, v in list_args if self._valid_to_convert(v)
            }

            converted_args = {
                k: (
                    (
                        self.convert(v)
                        if isinstance(v, dict) else
                        [self.convert(x) for x in v]
                    )
                ) for k, v in convertible_args.items()
            }

            # Merge with main JSON
            json = {**json, **converted_args}

        # Convert the main item, if possible
        spo_class = self._find_matching_class(json)

        if spo_class:
            return spo_class.from_dict(json)

        return json

    def _valid_to_convert(self, arg: Union[dict, list]) -> bool:
        if isinstance(arg, dict):
            return (
                arg.get("type") in types.SPOTIFY
                or arg.get("items")
                or self._find_matching_class(arg)
            )
        elif isinstance(arg, list):
            return all(self._valid_to_convert(v) for v in arg)
        else:
            return False

    def _find_matching_class(self, json: str) -> Optional[Type[types.BaseType]]:
        # Find matching class if possible, else try brute-forcing it
        spotify_json_type = json.get("type")
        if spotify_json_type in types.SPOTIFY.keys():
            possible_classes = [types.SPOTIFY.get(spotify_json_type)]
        else:
            possible_classes = [
                eval(f"types.{spo_class}") for spo_class in types.__all__[1:]
            ]

        # For each class, find if all JSON keys are in the class parameters
        for possible_class in possible_classes:
            class_args = inspect.signature(possible_class).parameters.items()
            if all(k in json or v.default is None for k, v in class_args):
                return possible_class
        else:
            return None



from __future__ import annotations
from typing import AsyncGenerator, Optional

import spotipy2
from spotipy2.types import Playlist, Paging, PlaylistItem


class PlaylistMethods:
    async def get_playlist(
        self: spotipy2.Spotify, playlist_id: str  # type: ignore
    ) -> Playlist:
        return await self._get(f"playlists/{self.get_id(playlist_id)}")

    async def get_playlist_tracks(
        self: spotipy2.Spotify,  # type: ignore
        playlist_id: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Paging:
        params = self.wrapper(market=market, limit=limit, offset=offset)

        playlist_tracks = await self._get(
            f"playlists/{self.get_id(playlist_id)}/tracks", params=params
        )

        return playlist_tracks

    async def iter_playlist_tracks(
        self: spotipy2.Spotify,  # type: ignore
        playlist_id: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: int = 0,
    ) -> AsyncGenerator[PlaylistItem]:
        while True:
            playlist_tracks = await self.get_playlist_tracks(
                playlist_id,
                market=market,
                limit=limit,
                offset=offset
            )

            for track in playlist_tracks.items:
                yield track

            offset += len(playlist_tracks.items)

            if not playlist_tracks.next:
                break
